#!/bin/sh
#    avrfuse - read and program AVR fuses using textual config files
#    Copyright (C) 2011  Tibor 'Igor2' Palinkas
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#    Author:
#     Tibor 'Igor2' Palinkas
#     avrfuse (at) igor2.repo.hu
#     http://repo.hu/projects/avrfuse


# Internals of the awk script
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Terminology:
#  - device map file describes possible fuse settings and default values (fuse.attiny2313 for example)
#  - config file is an actual set of setting for a project
#  - variables are the keys of key=value pairs in a config file
#  - section is a @section of the device map file
#  - item is a line of a section

# device map file image:
# DATA[sectiontype, section, item, fieldno]
#  sectiontype is "fuse" or "values"
#  section is fuse name or value name
#  item is unique item id (bit number for fuse, value pattern for values)
#  fieldno is field number or "found" for marking which items are already found
#
# cache:
#  ADDR_I2V[item] = variable:bit
#  ADDR_V2I[variable:bit] = item
#  (:bit is omitted for single-bit setting)
#
# current values
#  VALUE[variable] = numeric
#  stores values of variables in numeric form
#  !!! do not write this array directly, call set(key, value) instead !!!
#  !!! (that will update all related caches) !!!

short=0
verbose=0
sharedir="/usr/share/avrfuse"

while ! test -z "$1"
do
	case $1 in
		--read) op="read";;
		--write) op="write";;
		--default) op="default";;
		--dry) dry=1;;
		--short) short=$(($short+1)) ;;
		--verbose) verbose=$(($verbose+1)) ;;
		--start)
			shift 1
			start=$1
			;;
		--end)
			shift 1
			end=$1
			;;
		--prefix)
			shift 1
			prefix=$1
			;;
		--conf)
			shift 1
			iconf=$1
			oconf=$1
			;;
		--iconf)
			shift 1
			iconf=$1
			;;
		--oconf)
			shift 1
			oconf=$1
			;;
		-p)
			shift 1
			avr=$1
			;;
		*)
			dude="$dude $1"
	esac
	shift 1
done

awk -v "op=$op" -v "dude=$dude" -v "avr=$avr" -v "iconf=$iconf" -v "oconf=$oconf" -v "startconf=$start" -v "endconf=$end" -v "prefixconf=$prefix" -v "dry=$dry" -v "short=$short" -v "verbose=$verbose" -v "sharedir=$sharedir" '
BEGIN {
# all known fuse sections
	order="fuse hfuse lfuse efuse"
	ov = split(order, ORDER, " ")
	brace="{[^}]*}"
	q = "\""
	placeholder="\001"
}

# Strip leading/trailing whitespace from s and return stripped string
function strip(s)
{
	sub("^[ \t]*", "", s)
	sub("[ \t]*$", "", s)
	return s
}

# print error message to stderr and exit
function error(msg)
{
	print "avrfuse: " msg > "/dev/stderr"
	fatal=1
	exit 1
}

function tempfile_unsafe(       fn,tmp)
{
	print "WARNING: avrfuse: can not find mktemp or tempfile, have to create a temp file in an unsafe manner"
	srand()
	fn = "/tmp/avrfuse" rand()
	close(fn)
	if ((getline tmp < fn) >= 0)
		error("unsafe temp file " fn " already exists.")
	close(fn)
	print "" > fn
	close(fn)
	return fn
}

function tempfile(        cmd,fn)
{
	if (cmd == "") {
		fn = tempfile("mktemp")
		if (fn == "")
			fn = tempfile("tempfile")
		if (fn == "")
			fn = tempfile_unsafe();
		return fn;
	}
	close(cmd)
	cmd | getline fn
	close(cmd)
	return fn
}

# print configureation error and exit
function maperr(fn, lno, msg)
{
	error("Error loading device map " fn ":" lno " (" msg ")")
}

# load device map file
function load_map(fn         ,lno,fs,n,tmp,mode,midx,icol)
{
	close(fn)
	fs = FS;
	FS="[|]"
	for(lno = 0; (getline < fn) > 0;) {
		lno++
		# ignore comments
		if ($1 ~ "^[ \t]*#")
			continue
		tmp=strip($1)
		# identify section
		if (tmp == "@fuse") {
			mode = "fuse"
			midx = strip($2)
			icol = 2
			continue
		}
		else if (tmp == "@values") {
			mode  = "values"
			midx  = strip($2)
			icol  = 0
			items = 0
			HEADER[midx] = $3
			continue
		}
		# item lines should have at least two fields
		if (NF < 2)
			continue
		FOUND[mode, midx]++
		# index-column 0 means linear numbering under the section
		if (icol == 0) {
			item = items
			items++
			DATA[mode, midx, "items"] = items
		}
		else
			item = strip($icol)

		# allow one item per section with the same name
		if (DATA[mode, midx, item, "found"] != "")
			maperr(fn, lno, mode " " midx " has multiple definitions for item " item);

		# load each field in DATA
		for(n = 1; n <= NF; n++)
			DATA[mode, midx, item, n] = strip($n)
		ITEM[mode, midx, item] = 1
	}
	FS = fs;
	return lno;
}

# convert numeric to binary ("0b" format)
function conv2b(val)
{
	if (val ~ "^0b")
		return val
	if (val == int(val)) {
		error("TODO: conv2b for int " val)
	}
	error("conv2b: invalid numeric format")
}

# return a specific bit of a binary (0b) numeric
function getbit(bin, bit    ,b)
{
	if (!(bin ~ "^0b"))
		error("getbit called with non-binary value " bin)
	b = length(bin)-bit
	if (b < 2)
		error("getbit: invalid bit " bit " of " bin " which is too short")
	return substr(bin, b, 1)
}

# set a slot in VALUE and update all affected bits in all caches
function set(key, val     ,n,k,v)
{
	VALUE[key] = val
	for(n = 0; n < 16; n++) {
		k = key ":" n
		if (!(k in ADDR_V2I))
			break;
		if (v == "")
			v = conv2b(val)
		VALUE[ADDR_V2I[k]] = getbit(v, n)
	}
	update_values(0, 1)
}

# fill in V[prefix] with content of data field of "prefix{}" found in s
function getbrace(s, prefix, V)
{
	if (match(s, prefix brace)) {
		V[prefix] = substr(s, RSTART, RLENGTH);
		sub("^" prefix "{", "", V[prefix])
		sub("}$", "", V[prefix])
		sub(prefix brace, placeholder, s)
	}
	return s;
}

# report (map) interpreter error and exit
function error_interpret(msg, str)
{
	error("can not interpret settings: " msg " in string " q str q)
}

# interpret an item line, replacing the {} parts, collecting options in OPTS.
#  var is the variable of the item (first column in DATA[] for @fuse)
#  str is the string with {} blocks to interpret
#  OPTS is an array that will be filled in with options:
#   [numeric] = name
#   [len] = number of numerics
#   [setting] = @value name for v{} reference
#   [bit] = bit in @value for v{} reference
#  if setval is 1, create fake @value section for 1{}/0{} references and do not calculate actual return string much
function interpret(var, str, OPTS,      setval,    s,V,n)
{
	s = str
	s = getbrace(s, "0", V)
	s = getbrace(s, "1", V)
	s = getbrace(s, "f", V)
	s = getbrace(s, "v", V)
	if (("0" in V) || ("1" in V)) {
		if (!("0" in V))
			error_interpret("0{} is missing", str)
		if (!("1" in V))
			error_interpret("1{} is missing", str)
		if ("v" in V)
			error_interpret("v{} is provided together with 0{} and 1{}", str)
		if ("f" in V)
			error_interpret("f{} is provided together with 0{} and 1{}", str)

		if (setval) {
			DATA["values", var, "items"] = 2
			DATA["values", var, 0, 1] = 0
			DATA["values", var, 0, 2] = V["0"]
			DATA["values", var, 1, 1] = 1
			DATA["values", var, 1, 2] = V["1"]
		}
		else {
			OPTS["len"] = 2;
			OPTS[0] = V["0"] " (0)"
			OPTS[1] = V["1"] " (1)"
			sub(placeholder, V["0"] "/" V["1"], s)
			gsub(placeholder, "", s)
		}
		return s
	}
	if ("v" in V) {
		if ("1" in V)
			error_interpret("0{} is provided together with v{}", str)
		if ("0" in V)
			error_interpret("1{} is provided together with v{}", str)
		if ("f" in V)
			error_interpret("f{} is provided together with v{}", str)
		V["bit"] = V["v"]
		sub(":.*", "", V["v"])
		sub("^[^:]*:", "", V["bit"])
		OPTS["setting"] = V["v"]
		OPTS["bit"] = V["bit"]
		if (setval)
			return s

		sub(":.*", "", V["v"])
		OPTS["setting"] = V["v"]
		OPTS["len"] = 0
		for(n = 0; n < DATA["values", V["v"], "items"]; n++) {
			if (DATA["values", V["v"], n, "1"] != "!") {
				OPTS[n] = DATA["values", V["v"], n, "2"] " (" DATA["values", V["v"], n, "1"] ") " DATA["values", V["v"], n, "3"]
				OPTS["len"]++
			}
		}
		return HEADER[V["v"]]
	}
	return s
}

# print default config
function write_default(fn     ,f,n,var,SEEN,prfx)
{
	if (short < 1)
		prfx=prefixconf "#  "
	else
		prfx=""

	print startconf > fn
	# use the order of known fuse sections
	for(f = 1; f <= ov; f++) {
		# output section only if device map has it (not all device defines all fuse bytes)
		if (FOUND["fuse", ORDER[f]] != "") {
			if (short < 3)
				print prefixconf "# --- fuse byte " ORDER[f] > fn
			# check all 8 bits of the fuse byte
			for(n = 0; n < 8; n++) {
				var = DATA["fuse", ORDER[f], n, 1]
				if (var == "")
					continue;
				delete(OPTS)
				OPTS["setting"] = var
				tmp = interpret(var, DATA["fuse", ORDER[f], n, 3], OPTS, 0)
				# var may change in interpret when hitting a v{} reference
				var = OPTS["setting"]

				# same var may present multiple times in case of @value - print only once, when first found
				if (var in SEEN)
					continue
				SEEN[var]=1

				# print setting header in comment
				if (short < 3)
					printf("%s", prefixconf "# " tmp) > fn
				if (short < 1)
					print "" > fn
				else if (short < 2)
					printf(": ") > fn

				# print all options in comment
				for(o = 0; o < OPTS["len"]; o++) {
					if (short < 2)
						printf("%s", prfx OPTS[o]) > fn
					if ((short < 1) || (o == OPTS["len"]-1))
						print "" > fn
					else if (short < 2)
						printf("; ") > fn
				}

				# print actual setting
				print prefixconf var "=" lookup(var, VALUE[var]) > fn
			print prefixconf > fn
			}
			print prefixconf > fn
		}
	}
	print endconf > fn
}

# look up @value name for bits in var
function lookup(var, bits          ,n,v,ref)
{
	v = DATA["values", var, "items"]
	if (v == "") {
		error("lookup: " var " is not @values");
	}
	for(n = 0; n < v; n++) {
		ref = DATA["values", var, n, 1]
		if (ref == "!")
			return "INVALID VALUE: " bits
		if (ref == bits)
			return DATA["values", var, n, 2]
	}
}

# reverse lookup: look up bits for @value name for var
function rlookup(var, val          ,n,v,ref)
{
	v = DATA["values", var, "items"]
	if (v == "") {
		error("lookup: " var " is not @values");
	}
	for(n = 0; n < v; n++) {
		ref = DATA["values", var, n, 2]
		if (ref == "!")
			return "INVALID VALUE (A): " val
		if (ref == val)
			return DATA["values", var, n, 1]
	}
	return "INVALID VALUE (B): " val
}

# set device-map-default values in VALUE[]
function load_defaults(   n)
{
	for(n in ITEM)
		if (n ~ "^fuse")
			set(DATA[n, 1], DATA[n, 4])
}

# update composite @value from VALUE or device map defaults
#  var: name of composite variable
#  use_defaults: use device map defaults instead of VALUE
#  noset: do not call set() but directly set VALUE[] - useful only when called from set() to aboid infinite recursion
function value_update(var, use_defaults, noset,     b,bits,v)
{
	bits = ""
	for(b = 0; b < COMPMAX[var]; b++) {
		if (use_defaults)
			v = DATA[COMP[var, b], 4]
		else
			v = VALUE[DATA[COMP[var, b], 1]]
		bits=v bits
	}
	bits = "0b" bits
	if (noset)
		VALUE[var] = bits
	else
		set(var, bits)
}

# update VALUE for all composites
# use_defaults and noset are the same as invalue_update
function update_values(use_defaults, noset      ,n,var,V,SEEN)
{
	for(n in ITEM) {
		if (n ~ "^values") {
			split(n, V, "[" SUBSEP "]")
			var = V[2];
			if (var in SEEN)
				continue
			SEEN[var]=1
			value_update(var, use_defaults, noset);
		}
	}
}

# map all composites and set all composites using device map defaults
function load_composites_with_default(    n,var,V,SEEN,b,bits)
{
	for(n in ITEM)
		if (n ~ "^fuse")
			if (match(DATA[n, 3], "v" brace)) {
				name=substr(DATA[n, 3],RSTART, RLENGTH)
				sub("^v{", "", name)
				sub("}$", "", name)
				bit=name
				sub(":.*", "", name);
				sub(".*:", "", bit)
				COMP[name, bit] = n
				if (bit >= COMPMAX[name])
					COMPMAX[name] = bit+1
			}
	update_values(1, 0)
}

# map ADDR arrays
function setup_bitvals(     f,n,var,OPTS)
{
	for(f = 1; f <= ov; f++) {
		if (FOUND["fuse", ORDER[f]] != "") {
			for(n = 0; n < 8; n++) {
				var = DATA["fuse", ORDER[f], n, 1]
				if (var == "")
					continue;
				delete(OPTS)
				interpret(var, DATA["fuse", ORDER[f], n, 3], OPTS, 1)
				if (OPTS["bit"] != "") {
					ADDR_I2V[var] = OPTS["setting"] ":" OPTS["bit"]
					ADDR_V2I[OPTS["setting"] ":" OPTS["bit"]] = var
				}
				else {
					ADDR_I2V[var] = var
					ADDR_V2I[var] = var
				}
			}
		}
	}
}

# process fuse configuration: read fn and replace its settings using VALUE[]
#  fn is input config file name (mandatory)
#  to_fn is output config file name, which is:
#   - a different file name where result will be written to
#   - the same file name, in which case result is written in a temp file and them moved to fn
#   - empty string (no output is written but VALUES are read)
function proc_cfg(fn, to_fn        ,l,key,val,bits,mvcmd,mode,skip,prefixlen,SEEN)
{
	prefixlen = length(prefixconf)

	# if file names match, set up a temp file and a move command
	if (to_fn == fn) {
		to_fn = tempfile()
		mvcmd="mv " to_fn " " fn
	}
	else
		mvcmd = ""

	close(to_fn)
	close(fn)

	# mode 0 is before start, mode 2 is after end, mode 1 is in between (processing)
	# if there is no start sequence, assume processing from the start
	if (startconf == "")
		mode = 1
	else
		mode = 0

	for(l = 0; (getline < fn) > 0; l++) {
		# detect start and end sequences; when found, change mode and make sure the sequence line is echoed (skip)
		if ((startconf != "") && ($0 ~ startconf)) {
			mode = 1
			skip = 1
		}
		else if ((endconf != "") && ($0 ~ endconf)) {
			mode = 2
			skip = 1
		}


		# if there is a prefix and line does not start with prefix, echo it
		# or if it does start with prefix load key and val with the line,
		# prefix removed; NOTE: can not use regex for prefix processing
		# as prefix is not a regex but may contain regex control chars
		if ((mode == 1) && (prefixlen > 0)) {
			if (substr($0, 1, prefixlen) == prefixconf)
				key = substr($0, prefixlen, length($0))
			else
				skip = 1
		}
		else
			key = $0
		val = key

		# if line (without prefix) is comment or does not have "=" or we are
		# not processing or skipping single line, then echo and take next line
		if ((key ~ "^#") || !(key ~ "=") || (mode != 1) || skip) {
			skip = 0
			if (to_fn != "")
				print $0 > to_fn
			continue
		}

		# part before first "=" is key, rest is val
		sub("=.*", "", key)
		sub("^[^=]*=", "", val)

		if (key in SEEN)
			print "WARNING: multiple instances of variable " key " - please fix" > "/dev/stderr"
		SEEN[key] = 1

		# decide operation
		if (to_fn != "") {
			# write output - replace val with the corresponding value from VALUE
			val = lookup(key, VALUE[key])
			print prefixconf key "=" val > to_fn
		}
		else {
			# read-only; sanity checks on key and value
			if (!key in VALUE)
				error("Config file attempts to set key " key " which is not in device map " fn)
			bits = rlookup(key, val)
			if (bits ~ "^INVALID")
				error("Invalid value for key " key ": " val)

			# store value
			set(key, bits)
#			print "proc_cfg: key " = " bits " (" val ")" > "/dev/stderr"
		}
	}

	close(to_fn)
	close(fn)

	if (mvcmd != "")
		system(mvcmd)
}

# take VALUE[] and assamble a binary numeric ("0b") for the fuse section
function compile_fuse(fuse    ,var,bin)
{
	bin = ""
	for(n = 0; n < 8; n++) {
		var = DATA["fuse", fuse, n, 1]
		bin = VALUE[var] bin
	}
	return "0b" bin
}

# fill up binary numeric with 0s from the left to match len bits (len+2 total string length)
function left_adjust(bin, len,     n, v, s)
{
	if (!(bin ~ "0b"))
		error("left_adjust: not a 0b binary constant: " bin)
	v = len - (length(bin) - 2)
	if (v < 0)
		error("left_adjust: " bin " is too long to be adjusted to " len)
	if (v > 0) {
		s = "0"
		for(n = 1; n < v; n++)
			s = s "0"
	}
	sub("^0b", "0b" s, bin)
	return bin
}

# take numeric in bin and set VALUE bits of fuse accordingly
function load_fusebin(fuse, bin     ,n)
{
	bin = left_adjust(bin, 8)
	if (verbose)
		print "loading fuse byte: " fuse "=" bin
	for(n = 0; n < 8; n++)
		set(DATA["fuse", fuse, n, 1], getbit(bin, n))
}

function vprint(level, str)
{
	if (verbose >= level)
		print "avrfuse: " str > "/dev/stderr"
}

BEGIN {
	# argument sanity check
	if (avr == "")
		error("No avr device specified with -p");
	if (op == "")
		error("No operation specified; try --read, --write or --default")

	# load device map
	if ((load_map("fuse." avr) == 0) && (load_map(sharedir "/fuse." avr) == 0))
		error("Can not find fuse file for " avr)

	# set device map defaults and make all caches consistent
	load_defaults()
	load_composites_with_default()
	setup_bitvals();

	# input defaults to stdin, if input config file name is empty
	if ((iconf == "") || (iconf == "-"))
		iconf = "/dev/stdin"

	# output defaults to stdout, if output config file name is empty
	if ((oconf == "") || (oconf == "-"))
		oconf = "/dev/stdout"

	if (op == "default") {
		if (verbose)
			vprint(1, "Writing default " avr " fuse config to " oconf)
		write_default(oconf)
	}
	else if (op == "write") {
		# read config file, compile fuses and write to device using avrdude
		vprint(1, "Reading config file " iconf)
		proc_cfg(iconf, "")
		cmd = ""
		delete(TMP)
		vprint(1, "Compiling fuse bytes")
		for(f = 1; f <= ov; f++) {
			if (!(FOUND["fuse", ORDER[f]] != ""))
				continue
			fb = compile_fuse(ORDER[f])
			vprint(2, " fuse byte for " ORDER[f] " is " fb)
			cmd = cmd " -U " ORDER[f] ":w:" fb ":m"
		}
		cmd = "avrdude -p " avr " " cmd " " dude
		vprint(1, "Running avrdude to write fuses")
		vprint(2, " with command line " cmd)
		if (dry)
			print cmd
		else
			system(cmd)
	}
	else if (op == "read") {
		# read device using avrdude, read original config file and write output file
		cmd = ""
		l = 0;
		for(f = 1; f <= ov; f++) {
			if (!(FOUND["fuse", ORDER[f]] != ""))
				continue
			cmd = cmd " -U " ORDER[f] ":r:-:b"
			TMP[l] = ORDER[f]
			l++;
		}
		cmd = "avrdude -q -q -p " avr " " cmd " " dude
		close(cmd)
		l = 0;
		while((cmd | getline) > 0) {
			load_fusebin(TMP[l], $1)
			l++
		}
		close(cmd)
		proc_cfg(iconf, oconf)
	}
	vprint(1, "done.")
}

'
